use std::fmt;

use super::expr::Expr;
use super::unknown::*;
use super::Scalar;
use crate::math::expr::Expr::*;

#[derive(Clone, Debug, PartialEq)]
pub struct Eqn(pub Expr, pub Expr);

impl Unknowns for Eqn {
    fn unknowns(&self) -> UnknownSet {
        self.0
            .unknowns()
            .union(&self.1.unknowns())
            .cloned()
            .collect()
    }
    fn has_unknowns(&self) -> bool {
        self.0.has_unknowns() || self.1.has_unknowns()
    }
    fn has_unknown(&self, u: Unknown) -> bool {
        self.0.has_unknown(u) || self.1.has_unknown(u)
    }
}

fn ord_by_unkn(a: Expr, b: Expr, u: Unknown) -> Option<(Expr, Expr)> {
    if a.has_unknown(u) {
        Some((a, b))
    } else if b.has_unknown(u) {
        Some((b, a))
    } else {
        None
    }
}

impl Eqn {
    pub fn new(l: Expr, r: Expr) -> Self {
        Self(l, r)
    }

    pub fn simplify(self) -> Eqn {
        Eqn(self.0.simplify(), self.1.simplify())
    }

    pub fn solve(&self, for_u: Unknown) -> Option<Expr> {
        use Expr::*;
        if !self.has_unknown(for_u) {
            return None;
        }
        trace!("solve: {}", self);
        let (l, r) = (
            self.0
                .clone().distribute()
                .simplify(),
            self.1
                .clone().distribute()
                .simplify(),
        );
        let (mut l, mut r) = ord_by_unkn(l, r, for_u)?;
        loop {
            trace!("solve iter: {} == {}", l, r);
            let (new_l, new_r): (Expr, Expr) = match l {
                Unkn(u) => return if u == for_u { Some(r.simplify()) } else { None },
                Sum(es) => {
                    let (us, not_us): (Vec<_>, Vec<_>) =
                        es.into_iter().partition(|e| e.has_unknown(for_u));
                    if us.len() != 1 {
                        return None;
                    }
                    (
                        Sum(us).simplify(),
                        Expr::new_minus(r, Sum(not_us)).simplify(),
                    )
                }
                Product(es) => {
                    let (us, not_us): (Vec<_>, Vec<_>) =
                        es.into_iter().partition(|e| e.has_unknown(for_u));
                    if us.len() != 1 {
                        return None;
                    }
                    (
                        Product(us).simplify(),
                        Expr::new_div(r, Product(not_us)).simplify(),
                    )
                }
                Neg(v) => (*v, Expr::new_neg(r)),
                Div(num, den) => {
                    let (nu, du) = (num.has_unknown(for_u), den.has_unknown(for_u));
                    match (nu, du) {
                        (true, false) => (*num, Expr::new_product(r, *den)),
                        (false, true) => (Expr::new_product(r, *den), *num),
                        (true, true) => return None, // TODO: simplify
                        (false, false) => return None,
                    }
                }
                Const(_) => return None,
            };
            l = new_l;
            r = new_r;
        }
    }
}

impl fmt::Display for Eqn {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{} == {}", self.0, self.1)
    }
}

#[derive(Clone, Debug, PartialEq)]
pub struct Eqns(pub Vec<Eqn>);

impl Unknowns for Eqns {
    fn unknowns(&self) -> UnknownSet {
        self.0.iter().flat_map(|eqn: &Eqn| eqn.unknowns()).collect()
    }
    fn has_unknowns(&self) -> bool {
        self.0.iter().any(|eqn: &Eqn| eqn.has_unknowns())
    }
    fn has_unknown(&self, u: Unknown) -> bool {
        self.0.iter().any(|eqn: &Eqn| eqn.has_unknown(u))
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_unknowns() {
        let u1 = Unknown(1);
        let u2 = Unknown(2);
        let u3 = Unknown(3);
        assert!(u1.has_unknowns());
        assert!(u2.has_unknowns());
        assert!(u1.has_unknown(u1));
        assert!(!u1.has_unknown(u2));
        assert!(u1.unknowns().contains(&u1));
        assert!(!u2.unknowns().contains(&u1));

        let e1 = Expr::new_minus(Expr::Unkn(u1), Expr::Unkn(u2));
        assert!(e1.has_unknowns());
        assert!(e1.has_unknown(u1));
        assert!(e1.has_unknown(u2));
        assert!(!e1.has_unknown(u3));
        assert!(e1.unknowns().len() == 2);
    }

    fn const_expr(e: Expr) -> Option<Scalar> {
        match e {
            Expr::Const(c) => Some(c),
            _ => None,
        }
    }

    #[test]
    fn test_solve() {
        use Expr::*;
        let _ = env_logger::try_init();
        let u1 = Unknown(1);
        let e1 = Unkn(u1);
        let e2 = Const(1.);

        let eqn = Eqn(e1.clone(), e2.clone());
        assert_eq!(eqn.solve(u1), Some(Const(1.)));
        let eqn = Eqn(e2.clone(), e1.clone());
        assert_eq!(eqn.solve(u1), Some(Const(1.)));

        let e3: Expr = Expr::from(1.) + 1. + 2.;
        let eqn = Eqn(e1.clone(), e3.clone());
        assert_eq!(eqn.solve(u1), Some(Const(4.)));
        let e3 = Expr::from(1.) - 1.;
        let eqn = Eqn(e1.clone(), e3.clone());
        assert_eq!(eqn.solve(u1), Some(Const(0.)));

        let e1 = Expr::from(2.) / (Expr::from(1.) - 4.);
        let e2 = Expr::new_minus(Const(1.), Unkn(u1));
        let eqn = Eqn(e1, e2);
        info!("eqn: {} => {}", eqn, eqn.clone().simplify());
        let e = eqn.solve(u1).unwrap();
        assert!(const_expr(e.clone()).is_some());
        assert!(relative_eq!(const_expr(e.clone()).unwrap(), 5. / 3.));

        let e1 = Expr::from(2.) * (Expr::from(1.) - 4.);
        let e2 = (u1 * 2.) - u1;
        info!(
            "e1==e2: {}=={} => {}=={}",
            e1,
            e2,
            e1.clone().simplify(),
            e2.clone().simplify()
        );
        let eqn = Eqn(e1, e2);
        let e = eqn.solve(u1).unwrap();
        assert!(const_expr(e.clone()).is_some());
        assert!(relative_eq!(const_expr(e.clone()).unwrap(), -6.));

        let e1 = Expr::from(2.) * (Expr::from(1.) - 4.);
        let e2 = ((u1 * 2.) + (u1 * u1)) / u1;
        info!(
            "{}=={} distrib=> {}=={}",
            e1,
            e2,
            e1.clone().distribute(),
            e2.clone().distribute()
        );
        info!(
            "{}=={} simplify=> {}=={}",
            e1,
            e2,
            e1.clone().distribute().simplify(),
            e2.clone().distribute().simplify()
        );
        let eqn = Eqn(e1, e2);
        let e = eqn.solve(u1).unwrap();
        assert!(const_expr(e.clone()).is_some());
        assert!(relative_eq!(const_expr(e.clone()).unwrap(), -8.));

        let e1 = Expr::from(2.) * (Expr::from(1.) - (4.));
        let e2 = ((u1 * 2.) + (((u1 * u1) + u1) + (Expr::from(2.) - (1. + 1.)))) / u1;
        info!(
            "e1==e2: {}=={} => {}=={}",
            e1,
            e2,
            e1.clone().distribute().simplify(),
            e2.clone().distribute().simplify().simplify()
        );
        let eqn = Eqn(e1, e2);
        let e = eqn.solve(u1).unwrap();
        assert!(const_expr(e.clone()).is_some());
        assert!(relative_eq!(const_expr(e.clone()).unwrap(), -9.));
    }

}
